EPGP = {}

local player = {}
player.name = ""
player.ep = 0
player.gp = 0
player.characterNames = {}
player.recentLoot = {}

local nameMap = nil

local function CreateEmptyDatabase()
	local database2 = {}
	database2.version = 0.3
	database2.minEp = 0
	database2.maxEp = 2500
	database2.bonusEp = 100
	database2.bonusSeconds = 900
	database2.baseGp = 100
	database2.decayPercentage = 0.25
	database2.lastDecayDate = os.time()
	database2.decayDays = 7
	database2.standbyCheckFrequency = 30 * 60
	database2.standbyTimeout = 5 * 60
	database2.players = {}
	database2.audit = {}
	
	database2.itemLevelOffset = 1000
	
	database2.slotWeights = {
		["HighCost"] = 1,
		["MediumCost"] = 1,
		default = 1
	}
	
	database2.qualityWeights = {
		Epic = 0.75,
		Legendary = 1,
		default = 0.25
	}
	
	return database2
end

local database = CreateEmptyDatabase()

local function Audit(msg)
	if msg then
		local timeStr = os.date("%Y-%m-%d %H:%M:%S ", os.time())
		table.insert(database.audit, 1, timeStr .. msg)
		
		while #database.audit > 100 do
			table.remove(database.audit)
		end
	end
end

local function tolower(str)
	local ret = ""
	if str then
		for i=1,string.len(str) do
			local code = string.byte(str, i)
			if code < 65 or code > 90 then
				ret = ret .. string.char(code)
			else
				ret = ret .. string.lower(string.char(code))
			end
		end
	end
	
	return ret
end

local function RebuildNameMap()
	nameMap = {}
	for inName, player in pairs(database.players) do
		local name = tolower(inName)
		nameMap[name] = name
		
		for inCharacterName, _ in pairs(player.characterNames) do
			local characterName = tolower(inCharacterName)
			nameMap[characterName] = name
		end
	end
end

local function FindPlayerName(name)
	if not nameMap then
		RebuildNameMap()
	end
	
	local lowerName = tolower(name)
	return nameMap[lowerName]
end

local function IsNameUsed(name)
	local playerName = FindPlayerName(name)
	if playerName then
		local lowerName = tolower(name)
		if playerName == lowerName then
			return "account"
		else
			return "character"
		end
	end
	
	return nil
end

local function GetPlayer(inName)
	local playerName = FindPlayerName(inName)
	if playerName then
		return database.players[playerName], ""
	else
		return nil, "Could not find account " .. inName
	end	
end

local function AddPlayer(playerName)
	local used = IsNameUsed(playerName)
	if used then
		return false, "Name " .. playerName .. " is already used as a " .. used	
	end
	
	local player = {}
	player.name = playerName
	player.ep = 0
	player.gp = 0
	player.characterNames = {}
	player.recentLoot = {}
	local lowerName = tolower(playerName)
	database.players[lowerName] = player
	
	RebuildNameMap()

	local msg = "Account " .. playerName .. " added"
	Audit(msg)
	return true, msg
end

local function DeletePlayer(name)
	local playerName = FindPlayerName(name)
	if playerName then
		local lowerName = tolower(playerName)
		database.players[lowerName] = nil
		RebuildNameMap()
		local msg = "Account " .. playerName .. " deleted"
		Audit(msg)
		return true, msg
	else
		return false, "Could not find account " .. name
	end	
end

local function AddCharacter(playerName, characterName)
	local player, msg = GetPlayer(playerName)
	if not player then
		return false, msg
	end
	
	if not characterName then
		return false, "nil character name"
	end
	
	local lowerName = tolower(characterName)
	if player.characterNames[lowerName] then
		return true, "Character " .. characterName .. " already added"
	end
	
	player.characterNames[lowerName] = characterName
	
	RebuildNameMap()

	local msg = "Character " .. characterName .. " added to account " .. player.name
	Audit(msg)
	return true, msg
end

local function DeleteCharacter(playerName, characterName)
	local player, msg = GetPlayer(playerName)
	if not player then
		return false, msg
	end
	
	if not characterName then
		return false, "nil character name"
	end
	
	local lowerName = tolower(characterName)
	if not player.characterNames[lowerName] then
		return true, "Character " .. characterName .. " doesn't exist"
	end
	
	player.characterNames[lowerName] = nil

	RebuildNameMap()

	local msg = "Character " .. characterName .. " removed from account " .. player.name
	Audit(msg)
	return true, msg
end

local function GetWeight(weightTable, keyName)
	local value = weightTable[keyName]
	if not value then
		value = weightTable["default"]
	end
	
	return value
end

local function SetWeight(tableName, weightTable, inKeyName, value)
	local keyName = inKeyName
	if not weightTable[keyName] then
		keyName = "default"
	end
	
	local msg = ""
	if value and value ~= weightTable[keyName] then
		msg = "Changed " .. tableName .. " " .. inKeyName .. " from " .. weightTable[keyName] .. " to " .. value
		Audit(msg)
		weightTable[keyName] = value
	end
	
	return msg
end

local function GetSlotWeight(slotName)
	return GetWeight(database.slotWeights, slotName)
end

local function SetSlotWeight(slotName, value)
	return SetWeight("slot weight", database.slotWeights, slotName, value)
end

local function GetQualityWeight(qualityName)
	return GetWeight(database.qualityWeights, qualityName)
end

local function SetQualityWeight(qualityName, value)
	return SetWeight("quality weight", database.qualityWeights, qualityName, value)
end

local function GetSpecWeight(isMainSpec)
	if isMainSpec == true then
		return 1
	else
		return 4
	end
end

local function Round(value)
	return math.floor(value + 0.5)
end

local function CalculateItemValue(inLevel, qualityWeight)
	local itemLevel = inLevel - database.itemLevelOffset
	if itemLevel < 0 then
		itemLevel = 0
	end
	
	return itemLevel * qualityWeight
end

local function CalculateGearValue(slot, quality, itemLevel, specWeight)
	local slotWeight = GetSlotWeight(slot)
	local qualityWeight = GetQualityWeight(quality)
	local specWeight = GetSpecWeight(mainSpec)

	return Round(CalculateItemValue(itemLevel, qualityWeight) * 0.33 * slotWeight / specWeight);      --    SLOT x QUALITY x TIER / SPEC
end

local function AddBasicGp(playerName, lootName, value)
	local player = GetPlayer(playerName)
	if not player then
		return false, "Account not found"
	end
	
	player.gp = player.gp + value
	local msg = "Account " .. player.name .. " received item " .. lootName .. " for " .. value .. " gp"
	Audit(msg)
	return true, msg
end

local function AddGp(playerName, lootName, slot, quality, itemLevel, mainSpec)
	local player = GetPlayer(playerName)
	if not player then
		return false, "Account not found"
	end
	
	local value = CalculateGearValue(slot, quality, itemLevel, mainSpec)
	player.gp = player.gp + value
	local msg = "Account " .. player.name .. " received item " .. lootName .. " for " .. value .. " gp"
	Audit(msg)
	return true, msg
end

local function GetMaxEpLevel()
	local maxEp = 0
	for lowerName, player in pairs(database.players) do
		if player.ep > maxEp then
			maxEp = player.ep
		end
	end
	
	return maxEp
end

local function IsDecayNeeded()
	local ret = false
	
	if database.lastDecayDate and database.decayDays then
		local decayPeriod = database.decayDays * 86400
		if database.lastDecayDate + decayPeriod < os.time() then
			ret = true
		end
	end
	
	return ret
end

local function RescaleDatabase()
	local msg = ""
	local maxEp = GetMaxEpLevel()
	if maxEp > database.maxEp then
		local decayFactor = database.maxEp / maxEp
		if decayFactor > 0 then
			for lowerName, player in pairs(database.players) do
				player.ep = Round(player.ep * decayFactor)
				player.gp = Round(player.gp * decayFactor)			
			end
		end

		msg = string.format("Database scaled by %.2f%% due to max EP hit", decayFactor * 100)
		Audit(msg)
	end
	
	return msg
end

local function DecayDatabase()
	local msg = ""
	if database.decayPercentage <= 0 then
		msg = "Decay disabled, please make the decay percentage non-zero"
	else
		local decayFactor = 1 - database.decayPercentage
		if decayFactor > 0 then
			for lowerName, player in pairs(database.players) do
				player.ep = Round(player.ep * decayFactor)
				player.gp = Round((database.baseGp + player.gp) * decayFactor) - database.baseGp
				if player.gp < 0 then
					player.gp = 0
				end
			end
		end
		
		msg = string.format("Database decayed by %.2f%%", database.decayPercentage * 100)
		Audit(msg)	
		database.lastDecayDate = os.time()
	end

	return msg
end

local function AddEp(player, ep, reason)
	local player, msg = GetPlayer(player)
	if player then
		player.ep = player.ep + ep
		if ep > 0 then
			msg = "Account " .. player.name .. " ep increased by " .. ep .. " to " .. player.ep
		else
			msg = "Account " .. player.name .. " ep decreased by " .. (ep * -1) .. " to " .. player.ep
		end
		
		if reason and string.len(reason) > 0 then
			msg = msg .. ": " .. reason
		end
		
		Audit(msg)
		
		if player.ep > database.maxEp then
			msg = msg .. "\n" .. RescaleDatabase()
		end
	end
	
	return msg
end

local function GetPlayerCharacterNames(inPlayerName)
	local ret = {}
	local player = GetPlayer(inPlayerName)
	if player then
		for lowername, name in pairs(player.characterNames) do
			ret[lowername] = name
		end
	end	
	
	return ret
end

local function GetPlayerCharacterNamesAsString(player)
	if player then
		local ret = ""
		local firstName = true
		for lowername, name in pairs(player.characterNames) do
			if firstName then
				firstName = false
			else
				ret = ret .. ","
			end
			
			ret = ret .. name
		end
		
		return ret
	end	
	
	return ""
end

local function GetPlayerCharacterNamesAsJSONString(player)
	if player then
		local ret = ""
		local firstName = true
		for lowername, name in pairs(player.characterNames) do
			if firstName then
				firstName = false
			else
				ret = ret .. ", "
			end
			
			ret = ret .. "\"" .. name .. "\""
		end
		
		return ret
	end	
	
	return ""
end

local function GetPlayerEp(player)
	if player then
		return player.ep
	end
	
	return 0
end

local function GetPlayerGp(player)
	if player then
		return database.baseGp + player.gp
	end
	
	return 0
end

local function GetPlayerPriority(player)
	if player and player.ep > database.minEp then
		local totalGp = GetPlayerGp(player)
		if totalGp > 0 then
			return player.ep / totalGp
		end
	end
	
	return 0
end

local function GetPlayerNamesByPriority()
	local names = {}
	local priorities = {}
	local priorityArray = {}
	for lowerName, player in pairs(database.players) do
		local priority = GetPlayerPriority(player)
		local tempNames = priorities[priority]
		if not tempNames then
			tempNames = {}
			table.insert(tempNames, player.name)
			priorities[priority] = tempNames
			table.insert(priorityArray, priority)
		else
			table.insert(tempNames, player.name)
			table.sort(tempNames)
		end
	end
	
	table.sort(priorityArray, function(a, b) return a > b end)
	
	for _, priority in pairs(priorityArray) do
		local tempNames = priorities[priority]
		for _, name in ipairs(tempNames) do
			table.insert(names, name)
		end
	end
	
	return names
end

local function GetPlayerNames()
	local names = {}
	for lowerName, player in pairs(database.players) do
		table.insert(names, player.name)
	end
	
	table.sort(names)
	
	return names
end

local function GetBonusSeconds()
	return database.bonusSeconds
end

local function GetBonusEp()
	return database.bonusEp
end

local function GetMinEp()
	return database.minEp
end

local function GetMaxEp()
	return database.maxEp
end

local function GetBaseGp()
	return database.baseGp
end

local function GetDecayPercentage()
	return database.decayPercentage
end

local function GetItemLevelOffset()
	return database.itemLevelOffset
end

local function GetStandbyCheckFrequency()
	return database.standbyCheckFrequency
end

local function GetStandbyTimeout()
	return database.standbyTimeout
end

local function SetBonusSeconds(value)
	local msg = ""
	if value and value ~= database.bonusSeconds then
		msg = "Changed bonus seconds from " .. database.bonusSeconds .. " to " .. value
		Audit(msg)
		database.bonusSeconds = value
	end
	
	return msg
end

local function SetBonusEp(value)
	local msg = ""
	if value and value ~= database.bonusEp then
		msg = "Changed bonus ep from " .. database.bonusEp .. " to " .. value
		Audit(msg)
		database.bonusEp = value
	end
	
	return msg
end

local function SetMinEp(value)
	local msg = ""
	if value and value ~= database.minEp then
		msg = "Changed minimum EP from " .. database.minEp .. " to " .. value
		Audit(msg)
		database.minEp = value
	end
	
	return msg
end

local function SetMaxEp(value)
	local msg = ""
	if value and value ~= database.maxEp then
		msg = "Changed maximum EP from " .. database.maxEp .. " to " .. value
		Audit(msg)
		database.maxEp = value
		msg = msg .. RescaleDatabase()
	end
	
	return msg
end

local function SetBaseGp(value)
	local msg = ""
	if value and value ~= database.baseGp then
		msg = "Changed base GP from " .. database.baseGp .. " to " .. value
		Audit(msg)
		database.baseGp = value
	end
	
	return msg
end

local function SetDecayPercentage(value)
	local msg = ""
	if value and value ~= database.decayPercentage then
		msg = "Changed decay percentage from " .. database.decayPercentage .. " to " .. value
		Audit(msg)
		database.decayPercentage = value
	end
	
	return msg
end

local function SetItemLevelOffset(value)
	local msg = ""
	if value and value ~= database.itemLevelOffset then
		msg = "Changed item power offset from " .. database.itemLevelOffset .. " to " .. value
		Audit(msg)
		database.itemLevelOffset = value
	end
	
	return msg
end

local function GetDatabase()
	return database
end

local function SetDummyPlayer(name, ep, gp, altName)
	local player = GetPlayer(name)
	if not player then
		AddPlayer(name)
		local player = GetPlayer(name)
		player.ep = ep
		player.gp = gp
		AddCharacter(name, name)
		if altName then
			AddCharacter(name, altName)
		end
	end
end

local function SetDatabase(inDatabase)
	database = inDatabase
	
	if database.version == 0.1 then
		database.qualityWeights = nil
		database.slotWeights = nil
		database.itemLevelOffset = nil
	end
	
	if database.version < 0.3 then
		local tempMap = {}
		for name, player in pairs(database.players) do
			local lowerName = tolower(name)
			tempMap[name] = lowerName
		end
		
		for name, lowerName in pairs(tempMap) do
			local player = database.players[name]
			database.players[name] = nil
			database.players[lowerName] = player
			
			local tempCharMap = {}
			for characterName, _ in pairs(player.characterNames) do
				tempCharMap[characterName] = tolower(characterName)
			end
			
			for characterName, lowerCharacterName in pairs(tempCharMap) do
				player.characterNames[characterName] = nil
				player.characterNames[lowerCharacterName] = characterName
			end
		end
	end
	
	local tempDatabase = CreateEmptyDatabase()	
	for key, value in pairs(tempDatabase) do
		if not database[key] then
			database[key] = value
		end
	end
	
	database.version = tempDatabase.version
	
	RebuildNameMap()

	-- test data
	--SetDummyPlayer("Boorish", 1500, 0, "Boor")
	--SetDummyPlayer("Jesath", 1500, 0)
	--SetDummyPlayer("Player1", 1500, 0)
	--SetDummyPlayer("Player2", 500, 10)
	--SetDummyPlayer("Player3", 155, 8)
	--SetDummyPlayer("Player4", 1650, 90)
	--SetDummyPlayer("Player5", 1900, 10)
	--SetDummyPlayer("Player6", 600, 0)
	--SetDummyPlayer("Player7", 300, 0)
	--SetDummyPlayer("Player8", 800, 50)
	--SetDummyPlayer("Player9", 1500, 0)
	--SetDummyPlayer("Player10", 1500, 0)
	--SetDummyPlayer("Player11", 1500, 0)
	--SetDummyPlayer("Player12", 1500, 0)
	--SetDummyPlayer("Player13", 1500, 0)
	--SetDummyPlayer("Player14", 1500, 0)
	--SetDummyPlayer("Player15", 1500, 0)
end

local function GetAuditLogString()
	local auditStr = ""
	for i=1,#database.audit do
		auditStr = auditStr .. database.audit[i] .. "\n"
	end
	
	return auditStr
end

local function GetAuditLogJSONString()
	local auditStr = ""
	for i=1,#database.audit do
		if i > 1 then
			auditStr = auditStr .. ",\n"
		end
		
		auditStr = auditStr .. "\"" .. database.audit[i] .. "\""
	end
	
	auditStr = auditStr .. "\n"
	
	return auditStr
end

local function ExportDatabase()
	return Base64.Encode(serpent.dump(database))
end

local function ImportDatabase(importString)
	local newDatabase = serpent.load(Base64.Decode(importString))
	if newDatabase then
		SetDatabase(newDatabase)
	end
end

local function ExportDatabaseAsHTML()
	local html = "<table id=\"epgp\"><thead class=\"toprow\"><th class=\"toptextdata\">Account</th><th class=\"topdata\">EP</th><th class=\"topdata\">GP</th><th class=\"topdata\">Priority</th><th class=\"toptextdata\">Characters</th></thead>\n"
	for count, name in ipairs(GetPlayerNamesByPriority()) do
		local rowClass = "evenrow"
		if count % 2 == 1 then
			rowClass = "oddrow"
		end
		
		local player = GetPlayer(name)
		html = html .. "<tr class=\"" .. rowClass .. "\"><td class=\"textdata\">" .. name .. "</td><td class=\"data\">" .. GetPlayerEp(player) .. "</td><td class=\"data\">" .. GetPlayerGp(player) .. "</td><td class=\"data\">" .. string.format("%.1f", GetPlayerPriority(player)) .. "</td><td class=\"textdata\">" .. GetPlayerCharacterNamesAsString(player) .. "</td></tr>\n"
	end
	
	html = html .. "</table>\n"
	return html
end

local function ExportDatabaseAsCSV()
	local csv = "Account;EP;GP;Priority;Characters\n"
	for _, name in ipairs(GetPlayerNamesByPriority()) do
		local player = GetPlayer(name)
		csv = csv .. name .. ";" .. GetPlayerEp(player) .. ";" .. GetPlayerGp(player) .. ";" .. string.format("%.1f", GetPlayerPriority(player)) .. ";" .. GetPlayerCharacterNamesAsString(player) .. "\n"
	end

	return csv
end

local function ExportDatabaseAsBBCode()
	local bbCode = "[b]Account[/b]\n"
	for _, name in ipairs(GetPlayerNamesByPriority()) do
		local player = GetPlayer(name)
		bbCode = bbCode .. name .. " EP=" .. GetPlayerEp(player) .. " GP=" .. GetPlayerGp(player) .. " Priority=" .. string.format("%.1f", GetPlayerPriority(player)) .. " Chars=" .. GetPlayerCharacterNamesAsString(player) .. "\n"
	end
	
	bbCode = bbCode .. "\n[b]Audit Log[/b]\n" .. GetAuditLogString() .. "\n"
	bbCode = bbCode .. "\n[b]Exported database[/b]\n" .. ExportDatabase() .. "\n"
	return bbCode
end

local function ExportDatabaseAsJSON()
	local jsonCode = "{\n  \"accounts\": [\n"
	for count, name in ipairs(GetPlayerNamesByPriority()) do
		local player = GetPlayer(name)
		if count > 1 then
			jsonCode = jsonCode .. ",\n"
		end
		
		jsonCode = jsonCode .. "  {\n"
		jsonCode = jsonCode .. "    \"account\": \"" .. name .. "\",\n"
		jsonCode = jsonCode .. "    \"ep\": " .. GetPlayerEp(player) .. ",\n"
		jsonCode = jsonCode .. "    \"gp\": " .. GetPlayerGp(player) .. ",\n"
		jsonCode = jsonCode .. "    \"priority\": " .. string.format("%.1f", GetPlayerPriority(player)) .. ",\n"
		jsonCode = jsonCode .. "    \"characters\": [ " .. GetPlayerCharacterNamesAsJSONString(player) .. " ]\n"
		jsonCode = jsonCode .. "  }"
	end
	
	jsonCode = jsonCode .. "  ],\n"
	jsonCode = jsonCode .. "  \"audit\": [ " .. GetAuditLogJSONString() .. " ],\n"
	jsonCode = jsonCode .. "  \"database\": \"" .. ExportDatabase() .. "\"\n"
	jsonCode = jsonCode .. "}\n"
	return jsonCode
end

local function ExportDatabaseAsEnjin()
	local nameColumn = "[b]Account[/b]\n"
	local epColumn = "[right][b]EP[/b][/right]\n"
	local gpColumn = "[right][b]GP[/b][/right]\n"
	local priorityColumn = "[right][b]Priority[/b][/right]\n"
	local charactersColumn = "[b]Characters[/b]\n"
	for _, name in ipairs(GetPlayerNamesByPriority()) do
		local player = GetPlayer(name)
		nameColumn = nameColumn .. name .. "\n"
		epColumn = epColumn .. "[right]" .. GetPlayerEp(player) .. "[/right]\n"
		gpColumn = gpColumn .. "[right]" .. GetPlayerGp(player) .. "[/right]\n"
		priorityColumn = priorityColumn .. string.format("[right]%.1f[/right]\n", GetPlayerPriority(player))
		charactersColumn = charactersColumn .. GetPlayerCharacterNamesAsString(player) .. "\n"
	end
	
	local bbCode = "[columns]\n" .. nameColumn .. "[nextcol]\n" .. epColumn .. "[nextcol]\n" .. gpColumn .. "[nextcol]\n" .. priorityColumn .. "[nextcol]\n" .. charactersColumn .. "[/columns]\n"
	bbCode = bbCode .. "[br]\n[spoiler=Exported database]" .. ExportDatabase() .. "[/spoiler]\n"
	bbCode = bbCode .. "[br]\n[spoiler=Audit Log]" .. GetAuditLogString() .. "[/spoiler]\n"
	return bbCode
end

function EPGP.new()
	EPGP.GetPlayerNames = GetPlayerNames
	EPGP.GetPlayerNamesByPriority = GetPlayerNamesByPriority
	EPGP.FindPlayerName = FindPlayerName
	EPGP.GetPlayer = GetPlayer
	EPGP.GetPlayerEp = GetPlayerEp
	EPGP.GetPlayerGp = GetPlayerGp
	EPGP.GetPlayerPriority = GetPlayerPriority
	EPGP.GetPlayerCharacterNames = GetPlayerCharacterNames
	EPGP.GetPlayerCharacterNamesAsString = GetPlayerCharacterNamesAsString
	EPGP.AddPlayer = AddPlayer
	EPGP.DeletePlayer = DeletePlayer
	EPGP.AddCharacter = AddCharacter
	EPGP.DeleteCharacter = DeleteCharacter
	EPGP.GetDatabase = GetDatabase
	EPGP.SetDatabase = SetDatabase
	EPGP.ImportDatabase = ImportDatabase
	EPGP.ExportDatabase = ExportDatabase
	EPGP.ExportDatabaseAsBBCode = ExportDatabaseAsBBCode
	EPGP.ExportDatabaseAsEnjin = ExportDatabaseAsEnjin
	EPGP.ExportDatabaseAsHTML = ExportDatabaseAsHTML
	EPGP.ExportDatabaseAsCSV = ExportDatabaseAsCSV
	EPGP.ExportDatabaseAsJSON = ExportDatabaseAsJSON
	EPGP.GetAuditLogString = GetAuditLogString
	EPGP.GetBonusSeconds = GetBonusSeconds
	EPGP.GetBonusEp = GetBonusEp
	EPGP.GetMinEp = GetMinEp
	EPGP.GetMaxEp = GetMaxEp
	EPGP.GetBaseGp = GetBaseGp
	EPGP.GetDecayPercentage = GetDecayPercentage
	EPGP.GetItemLevelOffset = GetItemLevelOffset
	EPGP.GetSlotWeight = GetSlotWeight
	EPGP.GetQualityWeight = GetQualityWeight
	EPGP.GetStandbyCheckFrequency = GetStandbyCheckFrequency
	EPGP.GetStandbyTimeout = GetStandbyTimeout
	EPGP.SetBonusSeconds = SetBonusSeconds
	EPGP.SetBonusEp = SetBonusEp
	EPGP.SetMinEp = SetMinEp
	EPGP.SetMaxEp = SetMaxEp
	EPGP.SetBaseGp = SetBaseGp
	EPGP.SetDecayPercentage = SetDecayPercentage
	EPGP.SetItemLevelOffset = SetItemLevelOffset
	EPGP.SetSlotWeight = SetSlotWeight
	EPGP.SetQualityWeight = SetQualityWeight
	EPGP.AddEp = AddEp
	EPGP.AddGp = AddGp
	EPGP.AddBasicGp = AddBasicGp
	EPGP.CalculateGearValue = CalculateGearValue
	EPGP.DecayDatabase = DecayDatabase
	EPGP.IsDecayNeeded = IsDecayNeeded
	
	return EPGP
end
